import java.util.LinkedList;
import java.lang.reflect.Field;
import java.lang.management.*;
import sun.misc.Unsafe;

public class RatioVerifier {

    public static final long M = 1024 * 1024;
    public static final long MaxHeapSize = 200 * M;
    public static final long HeapSize =    10  * M;
    public static final long MaxNewSize =  20  * M;
    public static final long NewSize =     5   * M;

    private static Unsafe unsafe = null;

    static MemoryPoolMXBean bean = null;

    static {
	try {
	    Field theUnsafe = Unsafe.class.getDeclaredField("theUnsafe");
	    theUnsafe.setAccessible(true);
	    unsafe = (Unsafe) theUnsafe.get(null);
	} catch (NoSuchFieldException | IllegalAccessException e) {
	    throw new IllegalStateException(e);
	}

	for (MemoryPoolMXBean b : ManagementFactory.getMemoryPoolMXBeans()) {
	    if (b.getName().contains("Old") ||
		b.getName().contains("Tenured")) {
		bean = b;
		break;
	    }
	}

    }

    //Size of byte array that will be allocated
    public static final int CHUNK_SIZE = 1024;
    //Length of byte array, that will be added to "garbage" list.
    public static final int ARRAY_LENGTH = CHUNK_SIZE - Unsafe.ARRAY_BYTE_BASE_OFFSET;
    //Amount of tries to force heap shrinking/expansion using GC
    public static final int GC_TRIES = 10;

    //Value that will be added/substracted from expected min/max heap free ratio
    //during memory allocation to make sure that specified limit will be exceeded.
    public static final double OVERLOAD = 0.05;
    //Acceptable heap free ratio limit exceedance: verification will fail if
    //actual ratio is lower than expected min heap free ratio - VARIANCE or
    //higher than expected max heap free ratio + VARIANCE.
    public static final double VARIANCE = 0.025;

    public static LinkedList<Object> garbage = new LinkedList<Object>();

    public static void main(String args[]) throws Exception {

	double minRatio = Integer.valueOf(args[0])/100.0;
	double maxRatio = Integer.valueOf(args[1])/100.0;

	long maxHeapSize = getMax();

	long previouslyCommitted = 0;

	//commit 0.5 of total heap size to have enough space
	//to both shink and expand
	while (getCommitted() < maxHeapSize/2) {
	    garbage.add(new byte[ARRAY_LENGTH]);
	}

	forceGC();
	//Verify that current heap free ratio lies between specified limits
	verifyRatio(minRatio, maxRatio);

	//Estimeate how much memory we have to allocate to force expansion
	long memoryToFill = (long)(getCommitted()*(1-minRatio+OVERLOAD)) -
	    getUsed();

	previouslyCommitted = getCommitted();

	while (memoryToFill > 0) {
	    garbage.add(new byte[1024]);
	    memoryToFill -= CHUNK_SIZE;
	}

	forceGC();
	//Verify that after memory allocation heap free ratio is still conforming specified limits
	verifyRatio(minRatio, maxRatio);
	//Verify that heap was actually expanded
	if (previouslyCommitted >= getCommitted()) {
	    throw new RuntimeException("Heap was not expanded.");
	}

	//Estimate how much memory we have to free to force shrinking
	long memoryToFree = getUsed() -
	    (long)(getCommitted()*(1-maxRatio-OVERLOAD));

	previouslyCommitted = getCommitted();

	while (memoryToFree > 0 && garbage.size() > 0) {
	    garbage.remove(garbage.size() - 1);
	    memoryToFree -= CHUNK_SIZE;
	}

	forceGC();
	//Verify that heap free ratio is still conforming specified limits
	verifyRatio(minRatio, maxRatio);
	//Verify that heap was actually shrinked
	if (previouslyCommitted <= getCommitted()) {
	    throw new RuntimeException("Heap was not shrinked.");
	}

    }

    public static void forceGC() throws Exception {
	for (int i = 0; i < GC_TRIES; i++) {
	    System.gc();
	    Thread.sleep(10);
	}
    }

    public static void verifyRatio(double minRatio, double maxRatio) throws Exception {
	double ratio = getHeapFreeRatio();
	System.out.println(minRatio + " " + ratio + " " + maxRatio);
	if (minRatio - ratio > VARIANCE &&
	    getCommitted() < getMax()) {
	    throw new RuntimeException("Current heap free ratio is lower than "+
				       "MinHeapFreeRatio (" + ratio + " vs " + minRatio + ").");
	}
	if (ratio - maxRatio > VARIANCE &&
	    getUsed() > getInit()) {
	    throw new RuntimeException("Current heap free ratio is higher than "+
				       "MaxHeapFreeRatio (" + ratio + " vs " + maxRatio + ").");
	}
    }


    public static long getMax() {
	return bean.getUsage().getMax();
    }

    public static long getInit() {
	return bean.getUsage().getInit();
    }

    public static long getUsed() {
	return bean.getUsage().getUsed();
    }

    public static long getCommitted() {
	return bean.getUsage().getCommitted();
    }

    public static long getFree() {
	return getCommitted() - getUsed();
    }

    public static double getHeapFreeRatio() {
	return getFree()/(double)getCommitted();
    }

}
